package beanstalkd

import (
	"context"
	"flag"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("beanstalkd")

var IGNORE_METRICS = []string{
	"pid",
	"version",
	"rusage-utime",
	"rusage-stime",
	"id",
	"hostname",
}

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "beanstalkd.timeout", confTimeout, "[beanstalkd] operation timeout")
	fs.DurationVar(&confInterval, "beanstalkd.interval", confInterval, "[beanstalkd] interval between polls")
}

var Factory = newFactory()

type beanstalkdDaemonFactory struct {
	*daemon.TcpDaemonFactory
}

func newFactory() *beanstalkdDaemonFactory {
	tcpDaemonFactory := daemon.NewTcpDaemonFactory("beanstalkd").(*daemon.TcpDaemonFactory)

	return &beanstalkdDaemonFactory{
		TcpDaemonFactory: tcpDaemonFactory,
	}
}

func (df *beanstalkdDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	tcpDaemon := df.TcpDaemonFactory.New(paramsI).(*daemon.TcpDaemon)
	tcpDaemon.Timeout = confTimeout
	tcpDaemon.Interval_ = confInterval
	tcpDaemon.Keepalive = true
	tcpDaemon.Logger = log

	return &beanstalkdDaemon{
		TcpDaemon: tcpDaemon,
	}
}

type beanstalkdDaemon struct {
	*daemon.TcpDaemon
}

func (d *beanstalkdDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	if err := d.stats(ctx, "", m); err != nil {
		m.IncrError("beanstalkd", "beanstalkd")
		return err
	}

	tubes, err := d.listTubes(ctx)
	if err != nil {
		m.IncrError("beanstalkd", "beanstalkd")
		return err
	}

	var tubeErr error
	for _, tube := range tubes {
		if err := d.stats(ctx, tube, m); err != nil {
			m.IncrError("beanstalkd", "beanstalkd")
			tubeErr = err
		}
	}

	return tubeErr
}

func (d *beanstalkdDaemon) stats(ctx context.Context, tube string, m *common.MetricData) error {
	var cmd string
	if tube == "" {
		cmd = "stats"
		tube = "*"
	} else {
		cmd = "stats-tube " + tube
	}
	data, err := d.Command(ctx, cmd+"\r\n", []string{""})
	if err != nil {
		return fmt.Errorf("Error collecting beanstalkd '%s' metrics: %w", d.ServiceID, err)
	}

	for _, line := range data {
		kv := strings.SplitN(line, ": ", 2)
		if len(kv) != 2 {
			//TODO log
			continue
		}
		stat := kv[0]
		if common.Index(IGNORE_METRICS, stat) > -1 {
			continue
		}
		val, err := strconv.ParseFloat(kv[1], 64)
		if err != nil {
			//TODO log
			continue
		}
		labels := map[string]string{
			"namespace":    d.Namespace,
			"service":      d.ServiceID,
			"service_type": "beanstalkd",
			"service_addr": d.ServiceAddr,
		}

		// current-jobs-urgent, current-jobs-ready, current-jobs-reserved,
		// current-jobs-delayed, current-jobs-buried
		prefixJobs := "current-jobs-"
		prefixCmd := "cmd-"
		if strings.HasPrefix(stat, prefixJobs) {
			labels["status"] = stat[len(prefixJobs):]
			labels["tube"] = tube
			stat = "jobs"

			// cmd-put, cmd-peek, cmd-peek-ready, cmd-peek-delayed
			// cmd-peek-buried, cmd-reserve, cmd-reserve-with-timeout
			// cmd-delete, cmd-release, cmd-use, cmd-watch, cmd-ignore
			// cmd-bury, cmd-kick, cmd-touch, cmd-stats, cmd-stats-job
			// cmd-stats-tube, cmd-list-tubes, cmd-list-tube-used
			// cmd-list-tubes-watched, cmd-pause-tube
		} else if strings.HasPrefix(stat, prefixCmd) {
			labels["command"] = stat[len(prefixCmd):]
			stat = "commands_total"

		} else {
			switch stat {

			case "job-timeouts":
				stat = "job_timeouts_total"

			case "total-jobs":
				stat = "jobs_total"
				labels["tube"] = tube

			case "max-job-size":
				stat = "job_size_bytes_max"

			case "current-tubes":
				stat = "tubes"

			case "current-connections":
				stat = "connections"
				labels["type"] = "open"
				labels["tube"] = tube
			case "current-producers":
				stat = "connections"
				labels["type"] = "producer"
				labels["tube"] = tube
			case "current-workers":
				stat = "connections"
				labels["type"] = "worker"
				labels["tube"] = tube
			case "current-waiting":
				stat = "connections"
				labels["type"] = "waiting"
				labels["tube"] = tube
			case "current-using":
				stat = "connections"
				labels["type"] = "using"
				labels["tube"] = tube
			case "current-watching":
				stat = "connections"
				labels["type"] = "watching"
				labels["tube"] = tube

			case "total-connections":
				stat = "connections_total"

			case "uptime":
				stat = "uptime_seconds"

			case "binlog-records-migrated":
				stat = "binlog_records"
				labels["type"] = "migrated"
			case "binlog-records-written":
				stat = "binlog_records"
				labels["type"] = "written"

			case "binlog-max-size":
				stat = "binlog_size_bytes_max"

			case "pause":
				stat = "tube_pause_seconds"
				labels["tube"] = tube
			case "pause-time-left":
				stat = "tube_pause_left_seconds"
				labels["tube"] = tube

			/* by default use the metrics as-is and consider it's a float */
			default:
				/* non exhaustive list:
				 *   binlog-oldest-index
				 *   binlog-current-index
				 */
				stat = strings.ReplaceAll(stat, "-", "_")
			}
		}
		m.Set(&common.Metric{
			Name:   "beanstalkd_" + stat,
			Labels: labels,
			Value:  val,
			TTL:    confInterval * 2,
		})
	}

	return nil
}

func (d *beanstalkdDaemon) listTubes(ctx context.Context) ([]string, error) {
	data, err := d.Command(ctx, "list-tubes\r\n", []string{""})
	if err != nil {
		return nil, fmt.Errorf("Error listing beanstalkd '%s' tubes: %w", d.ServiceID, err)
	}

	tubes := make([]string, 0, len(data))
	for _, line := range data {
		kv := strings.SplitN(line, " ", 2)
		if len(kv) != 2 || kv[0] != "-" || kv[1] == "" {
			continue
		}
		tubes = append(tubes, kv[1])
	}

	return tubes, nil
}
